from __future__ import annotations
from collections import ChainMap
from dataclasses import dataclass
from typing import Optional, Tuple

from .LoopIR import LoopIR, T, LoopIR_Compare
from .new_eff import Check_ExprBound
from .prelude import Sym, _null_srcinfo_obj


# TODO: we should implement a more general index analysis which
# will leverage SMT solver to prove comparisons. But we should still
# keep all the code below for fast, constant bounds inference.


def binop(op: str, e1: LoopIR.expr, e2: LoopIR.expr):
    return LoopIR.BinOp(op, e1, e2, e1.type, e1.srcinfo)


def zero():
    return LoopIR.Const(0, T.index, _null_srcinfo_obj)


def is_zero(e):
    return isinstance(e, LoopIR.Const) and e.val == 0


@dataclass
class IndexRange:
    """
    Represents a range of possible values between [base + lo, base + hi]
        - [base] contains all expressions that we don't want to bounds infer (e.g. free
        variables.
        - [lo] and [hi] are bounds. If either is None, it means that there is no
        constant bound
    """

    base: LoopIR.expr
    lo: Optional[int]
    hi: Optional[int]

    def create_unbounded() -> IndexRange:
        return IndexRange(zero(), None, None)

    def create_int(x: int) -> IndexRange:
        return IndexRange(zero(), x, x)

    def create_constant_range(lo: int, hi: int) -> IndexRange:
        return IndexRange(zero(), lo, hi)

    def get_bounds(self) -> Tuple[str, str]:
        """
        Returns a tuple of the form (base + lo, base + hi + 1). If either
        endpoint is unboundable, returns inf or -inf.
        """
        if is_zero(self.base):
            lo = "-inf" if self.lo is None else str(self.lo)
            hi = "inf" if self.hi is None else str(self.hi)
        else:
            lo = "-inf" if self.lo is None else f"{self.base} + {self.lo}"
            hi = "inf" if self.hi is None else f"{self.base} + {self.hi + 1}"
        return lo, hi

    def get_stride_of(self, idx: Sym) -> int:
        assert isinstance(idx, Sym)

        def get_coeff(e):
            """
            This implementation assumes that self.base is a linear combination of index
            variables WITHOUT a constant term, e.g. ax + by + cz where a, b, c are constants
            and x, y, z are index expression. This holds because constant terms are always
            folded into self.lo and self.hi
            """
            if isinstance(e, LoopIR.Read):
                return 1 if e.name == idx else 0
            elif isinstance(e, LoopIR.Const):
                return e.val
            elif isinstance(e, LoopIR.BinOp):
                lhs = get_coeff(e.lhs)
                rhs = get_coeff(e.rhs)
                op = e.op
                if op == "+":
                    return lhs + rhs
                elif op == "-":
                    return lhs - rhs
                elif op == "*":
                    return lhs * rhs
                else:
                    raise ValueError(
                        f"cannot get stride of {idx} because {e} contains an unsupported operand '{op}'"
                    )
            elif isinstance(e, LoopIR.USub):
                return -get_coeff(e.arg)
            return 0

        return get_coeff(self.base)

    def partial_eval_with_range(self, var: Sym, rng: IndexRange) -> IndexRange:
        c = self.get_stride_of(var)
        if c == 0:
            return self

        new_bounds = index_range_analysis(self.base, {var: (rng.lo, rng.hi)})

        if is_zero(rng.base):
            return new_bounds
        else:
            c_expr = LoopIR.Const(c, T.index, self.base.srcinfo)
            new_base = LoopIR.BinOp("*", c_expr, rng.base, T.index, self.base.srcinfo)
            return new_bounds + IndexRange(new_base, 0, 0)

    def get_size(self) -> int | None:
        if self.lo is None or self.hi is None:
            return None
        return self.hi - self.lo + 1  # +1 because bounds are inclusive

    def __str__(self) -> str:
        base = "0" if is_zero(self.base) else str(self.base)
        lo = "-inf" if self.lo is None else str(self.lo)
        hi = "inf" if self.hi is None else str(self.hi)
        return f"({base}, {lo}, {hi})"

    def __add__(self, other: int | IndexRange) -> IndexRange:
        assert isinstance(other, (int, IndexRange))
        if isinstance(other, int):
            new_lo, new_hi = None, None
            if self.lo is not None:
                new_lo = self.lo + other
            if self.hi is not None:
                new_hi = self.hi + other
            return IndexRange(self.base, new_lo, new_hi)
        else:
            if is_zero(self.base):
                new_base = other.base
            elif is_zero(other.base):
                new_base = self.base
            else:
                new_base = binop("+", self.base, other.base)

            new_lo, new_hi = None, None
            if self.lo is not None and other.lo is not None:
                new_lo = self.lo + other.lo
            if self.hi is not None and other.hi is not None:
                new_hi = self.hi + other.hi
            return IndexRange(new_base, new_lo, new_hi)

    def __radd__(self, c: int) -> IndexRange:
        assert isinstance(c, int)
        return self.__add__(c)

    def __neg__(self) -> IndexRange:
        new_base, new_lo, new_hi = zero(), None, None
        if not is_zero(self.base):
            new_base = LoopIR.USub(self.base, self.base.type, self.base.srcinfo)
        if self.lo is not None:
            new_hi = -self.lo
        if self.hi is not None:
            new_lo = -self.hi
        return IndexRange(new_base, new_lo, new_hi)

    def __sub__(self, other: int | IndexRange) -> IndexRange:
        assert isinstance(other, (int, IndexRange))
        return self + (-other)

    def __rsub__(self, c: int) -> IndexRange:
        assert isinstance(c, int)
        return -self + c

    def __mul__(self, c: int) -> IndexRange:
        assert isinstance(c, int)
        if c == 0:
            return 0

        new_base, new_lo, new_hi = zero(), None, None
        if not is_zero(self.base):
            const = LoopIR.Const(c, T.index, self.base.srcinfo)
            new_base = binop("*", self.base, const)
        if self.lo is not None:
            new_lo = self.lo * c
        if self.hi is not None:
            new_hi = self.hi * c

        if c > 0:
            return IndexRange(new_base, new_lo, new_hi)
        else:
            return IndexRange(new_base, new_hi, new_lo)

    def __rmul__(self, c: int) -> IndexRange:
        assert isinstance(c, int)
        return self.__mul__(c)

    def __floordiv__(self, c: int) -> IndexRange:
        assert isinstance(c, int)
        if c == 0:
            return ValueError("Cannot divide by 0.")
        elif c < 0:
            return IndexRange.create_unbounded()

        new_lo, new_hi = None, None
        if is_zero(self.base):
            if self.lo is not None:
                new_lo = self.lo // c
            if self.hi is not None:
                new_hi = self.hi // c

            return IndexRange.create_constant_range(new_lo, new_hi)
        else:
            # TODO: Maybe can do some reasoning about base and lo if new_lo is not None
            c_expr = LoopIR.Const(c, T.index, self.base.srcinfo)
            new_base = LoopIR.BinOp(
                "/", self.base, c_expr, self.base.type, self.base.srcinfo
            )
            if self.lo is not None and self.hi is not None:
                new_lo = self.lo // c
                new_hi = self.hi // c + 1

            return IndexRange(new_base, new_lo, new_hi)

    def __mod__(self, c: int) -> IndexRange:
        assert isinstance(c, int)
        if (
            is_zero(self.base)
            and self.lo is not None
            and self.hi is not None
            and self.lo // c == self.hi // c
        ):
            return IndexRange.create_constant_range(self.lo % c, self.hi % c)
        return IndexRange.create_constant_range(0, c - 1)

    # join
    def __or__(self, other: IndexRange) -> IndexRange:
        assert isinstance(other, IndexRange)
        compare_ir = LoopIR_Compare()
        if (is_zero(self.base) and other.base is None) or compare_ir.match_e(
            self.base, other.base
        ):
            if self.lo is None:
                new_lo = other.lo
            elif other.lo is None:
                new_lo = self.lo
            else:
                new_lo = min(self.lo, other.lo)
            if self.hi is None:
                new_hi = other.hi
            elif other.hi is None:
                new_hi = self.hi
            else:
                new_hi = max(self.hi, other.hi)
            return IndexRange(self.base, new_lo, new_hi)
        else:
            return IndexRange.create_unbounded()


def index_range_analysis(
    expr: LoopIR.expr, env: ChainMap | dict = {}
) -> IndexRange | int:
    """
    Based on the supplied [env], recursively performs range analysis on
    the possible values for [expr]. Returns either an int if the value
    is constant, or an IndexRange of the form [offset+lo, offset+hi].

    When env is empty, this function effectively just separates the
    constants from the non-constants index expressions.
    """
    assert isinstance(expr, LoopIR.expr)
    assert isinstance(env, (ChainMap, dict))

    def analyze_range(expr) -> IndexRange | int:
        assert isinstance(expr, LoopIR.expr)

        if not expr.type.is_indexable():
            raise ValueError("Cannot analyze range of non-index expression")

        if isinstance(expr, LoopIR.Read):
            sym = expr.name
            if sym not in env:
                return IndexRange(expr, 0, 0)
            lo, hi = env[sym]
            return IndexRange.create_constant_range(lo, hi)
        elif isinstance(expr, LoopIR.Const):
            return expr.val
        elif isinstance(expr, LoopIR.USub):
            return -analyze_range(expr.arg)
        elif isinstance(expr, LoopIR.BinOp):
            lhs_range = analyze_range(expr.lhs)
            rhs_range = analyze_range(expr.rhs)
            if expr.op == "+":
                return lhs_range + rhs_range
            elif expr.op == "-":
                return lhs_range - rhs_range
            elif expr.op == "*":
                return lhs_range * rhs_range
            elif expr.op == "/":
                return lhs_range // rhs_range
            elif expr.op == "%":
                return lhs_range % rhs_range
            else:
                assert False, "invalid binop in index expression"
        else:
            assert False, "invalid expr in index expression"

    return analyze_range(expr)


def constant_bound(expr, env) -> Tuple[int, int] | None:
    """
    Returns constant integer bounds for [expr], if possible, and
    None otherwise. The bounds are inclusive.
    """
    if isinstance(expr, int):
        return (expr, expr)

    idx_rng = index_range_analysis(expr, env)
    if isinstance(idx_rng, int):
        return (idx_rng, idx_rng)

    if not is_zero(idx_rng.base):
        return (None, None)
    return (idx_rng.lo, idx_rng.hi)


def arg_range_analysis(proc, arg, fast=True):
    """
    Try to find a bounding range on the arguments
    of a proc.

    Returns: A tuple `T` of length 2 s.t. it represents the range
    `[T[0], T[1]]` (both inclusive).
    If `T[0]` or `T[1]` is `None` it represents no knowledge
    of the value of that side of the range.

    NOTE: This is analysis can be slow. We should
    consider doing it when constructing a new proc
    and storing the result within LoopIR.fnarg.
    """
    assert arg.type.is_indexable()

    if fast:
        if isinstance(arg.type, LoopIR.Size):
            return (1, None)
        else:
            return (None, None)

    def lower_bound_check(value):
        return Check_ExprBound(
            proc,
            [proc.body[0]],
            LoopIR.Read(name=arg.name, idx=[], type=T.size, srcinfo=proc.srcinfo),
            ">=",
            value,
            exception=False,
        )

    def upper_bound_check(value):
        return Check_ExprBound(
            proc,
            [proc.body[0]],
            LoopIR.Read(name=arg.name, idx=[], type=T.size, srcinfo=proc.srcinfo),
            "<=",
            value,
            exception=False,
        )

    def binary_search_lower_bound(left, right):
        result = None

        while left <= right:
            mid = left + (right - left) // 2
            if lower_bound_check(mid):
                result = mid
                left = mid + 1
            else:
                right = mid - 1

        return result

    def binary_search_upper_bound(left, right):
        result = None

        while left <= right:
            mid = left + (right - left) // 2
            if upper_bound_check(mid):
                result = mid
                right = mid - 1
            else:
                left = mid + 1

        return result

    # Let's try to find a bounding range on args.
    # The upper bound on the absolute value of the range is a
    # reasonable large value so that if an answer exists
    # it will probably be within this.
    max_abs_search = 2**15

    min_search = 1 if isinstance(arg.type, LoopIR.Size) else -max_abs_search

    lower_bound = binary_search_lower_bound(min_search, max_abs_search)
    upper_bound = binary_search_upper_bound(min_search, max_abs_search)

    return (lower_bound, upper_bound)


class IndexRangeEnvironment:
    lt = "<"
    leq = "<="
    eq = "=="

    @staticmethod
    def get_pred_reads(expr):
        if isinstance(expr, LoopIR.Read):
            return {expr.name}
        elif isinstance(expr, LoopIR.USub):
            return IndexRangeEnvironment.get_pred_reads(expr.arg)
        elif isinstance(expr, LoopIR.BinOp):
            return IndexRangeEnvironment.get_pred_reads(
                expr.lhs
            ) | IndexRangeEnvironment.get_pred_reads(expr.rhs)
        else:
            return set()

    def __init__(self, proc, fast=True) -> None:
        assert isinstance(proc, LoopIR.proc)

        preds_reads = set()
        if not fast:
            # Get parameters referenced in predicates
            # to only analyze those
            for pred in proc.preds:
                preds_reads = preds_reads | IndexRangeEnvironment.get_pred_reads(pred)

        self.proc = proc
        self.env = ChainMap()
        for arg in proc.args:
            if isinstance(arg.type, LoopIR.Size):
                self.env[arg.name] = arg_range_analysis(
                    proc, arg, fast=arg.name not in preds_reads
                )

    def enter_scope(self):
        self.env = self.env.new_child()

    def exit_scope(self):
        self.env = self.env.parents

    def add_loop_iter(self, sym, lo_expr, hi_expr):
        lo, _ = constant_bound(lo_expr, self.env)
        _, hi = constant_bound(hi_expr, self.env)
        if hi is not None:
            hi = hi - 1

        sym_range = (lo, hi)

        # This means that this variable's range is invalid e.g.
        # `for i in seq(4, 4)`
        if (
            sym_range[0] is not None
            and sym_range[1] is not None
            and sym_range[0] > sym_range[1]
        ):
            # TODO: this probably shouldn't get added as (None, None), since that
            # means it could take on any possible value.
            sym_range = (None, None)

        self.env[sym] = sym_range

    @staticmethod
    def _check_range(range0, op, range1):
        if range0[1] is None or range1[0] is None:
            return False
        if op == IndexRangeEnvironment.lt:
            return range0[1] < range1[0]
        elif op == IndexRangeEnvironment.leq:
            return range0[1] <= range1[0]
        else:
            if range0[0] is None or range1[1] is None:
                return False
            return range0[0] == range0[1] == range1[0] == range1[1]

    def check_expr_bound(self, expr0, op, expr1):
        expr0_range = constant_bound(expr0, self.env)
        expr1_range = constant_bound(expr1, self.env)
        return IndexRangeEnvironment._check_range(expr0_range, op, expr1_range)

    def check_expr_bounds(self, expr0, op0, expr1, op1, expr2):
        expr0_range = constant_bound(expr0, self.env)
        expr1_range = constant_bound(expr1, self.env)
        expr2_range = constant_bound(expr2, self.env)
        return IndexRangeEnvironment._check_range(
            expr0_range, op0, expr1_range
        ) and IndexRangeEnvironment._check_range(expr1_range, op1, expr2_range)



from __future__ import annotations

import pytest

from exo import proc, ExoType
from exo.libs.memories import *
from exo.API_cursors import *

from exo.stdlib.inspection import *
from exo.stdlib.scheduling import *


@pytest.fixture(scope="session")
def proc_foo():
    @proc
    def foo(n: size, m: size):
        for i in seq(0, n):
            for j in seq(0, m):
                x: f32
                x = 0.0
                y: f32
                y = 1.1

    yield foo


@pytest.fixture(scope="session")
def proc_bar():
    @proc
    def bar(n: size, m: size):
        x: f32
        for i in seq(0, n):
            for j in seq(0, m):
                x = 0.0
                x = 1.0
                x = 2.0
                x = 3.0
                x = 4.0
                x = 5.0

    yield bar


def test_parent_cursor(proc_foo):
    jloop = proc_foo.find("for j in _:_")
    iloop = proc_foo.find("for i in _:_")
    assert jloop.parent() == iloop


def test_child_cursor(proc_foo):
    jloop = proc_foo.find("for j in _:_")
    iloop = proc_foo.find("for i in _:_")
    assert jloop == iloop.body()[0]


def test_asblock_cursor(proc_foo):
    jloop = proc_foo.find("for j in _:_")
    iloop = proc_foo.find("for i in _:_")
    assert jloop.as_block() == iloop.body()


def test_builtin_cursor():
    @proc
    def bar():
        x: f32
        x = select(0.0, 1.0, 2.0, 3.0)

    select_builtin = bar.find("select(_)")
    assert select_builtin == bar.body()[1].rhs()


def test_basic_forwarding(golden):
    @proc
    def p():
        x: f32
        if True:
            x = 1.0
        if True:
            x = 2.0

    x1 = p.find("x = _ #0")
    x2 = p.find("x = _ #1")
    if1 = x1.parent()
    if2 = x2.parent()
    p = insert_pass(p, if1.before())
    p = fuse(p, if1, if2)
    assert str(p) == golden


def test_gap_forwarding(golden):
    @proc
    def p():
        x: f32
        if True:
            x = 1.0
        if True:
            x = 2.0

    if1 = p.find("x = _ #0").parent()
    if2 = p.find("x = _ #1").parent()
    x_alloc = p.find("x: _")
    p = fuse(p, if1, if2)
    p = insert_pass(p, if1.body()[0].after())
    p = insert_pass(p, if2.body()[0].after())
    p = insert_pass(p, x_alloc.after())
    assert str(p) == golden


def test_basic_forwarding2(golden):
    @proc
    def filter1D(ow: size, kw: size, x: f32[ow + kw - 1], y: f32[ow], w: f32[kw]):
        for o in seq(0, ow):
            sum: f32
            sum = 0.0
            for k in seq(0, kw):
                sum += x[o + k] * w[k]
            y[o] = sum

    filter1D = divide_loop(filter1D, "o", 4, ["outXo", "outXi"], tail="cut_and_guard")

    sum_c = filter1D.find("sum:_")

    filter1D = expand_dim(filter1D, sum_c, "4", "outXi")
    filter1D = lift_alloc(filter1D, sum_c)

    assert str(filter1D.forward(sum_c)) == golden


def test_basic_forwarding3(golden):
    @proc
    def filter1D(ow: size, kw: size, x: f32[ow + kw - 1], y: f32[ow], w: f32[kw]):
        for o in seq(0, ow):
            sum: f32
            sum = 0.0
            for k in seq(0, kw):
                sum += x[o + k] * w[k]
            y[o] = sum

    filter1D = divide_loop(filter1D, "o", 4, ["outXo", "outXi"], tail="cut_and_guard")

    sum_c = filter1D.find("sum:_")

    filter1D = expand_dim(filter1D, sum_c, "4", "outXi")
    filter1D = lift_alloc(filter1D, filter1D.forward(sum_c))

    assert str(filter1D.forward(sum_c)) == golden


def test_simplify_forwarding(golden):
    @proc
    def foo(n: size, m: size):
        x: R[n, 16 * (n + 1) - n * 16, (10 + 2) * m - m * 12 + 10]
        for i in seq(0, 4 * (n + 2) - n * 4 + n * 5):
            y: R[n + m]
            y[n * 4 - n * 4 + 1] = 0.0

    stmt = foo.find("y[_] = _")
    foo1 = simplify(foo)
    assert str(foo.find("y:_").shape()[0]._impl._node) == "n + m"
    assert str(foo1.forward(stmt)._impl._node) == golden


def test_simplify_predicates_forwarding():
    @proc
    def foo(n: size):
        assert n >= 1
        for i in seq(0, n):
            pass

    loop = foo.find_loop("i")
    foo = simplify(foo)
    loop = foo.forward(loop)
    assert loop == foo.find_loop("i")


def test_type_and_shape_introspection():
    @proc
    def foo(n: size, m: index, flag: bool):
        assert m >= 0
        a: R[n + m]
        b: i32[n]
        c: i8[n]
        d: f32[n]
        e: f64[2]

    assert foo.find("a:_").type() == ExoType.R
    assert foo.find("b:_").type() == ExoType.I32
    assert foo.find("c:_").type() == ExoType.I8
    assert foo.find("d:_").type() == ExoType.F32
    assert foo.find("e:_").type() == ExoType.F64
    assert foo.args()[0].type() == ExoType.Size
    assert foo.args()[1].type() == ExoType.Index
    assert foo.args()[2].type() == ExoType.Bool
    assert str(foo.find("a:_").shape()[0]._impl._node) == "n + m"
    assert foo.find("e : _").shape()[0].type() == ExoType.Int


def test_expand_dim_forwarding(golden):
    @proc
    def scal(n: size, alpha: R, x: [R][n]):
        for i in seq(0, n):
            x[i] = alpha * x[i]

    stmt = scal.find("x[_] = _")
    scal1 = divide_loop(scal, "for i in _:_", 8, ("io", "ii"), tail="cut")
    scal2 = bind_expr(scal1, [stmt.rhs().lhs()], "alphaReg")

    # seems to fail for some other scheduling ops besides expand_dim...
    # scal3 = lift_alloc(scal2, scal2.find("alphaReg: _"))
    scal3 = expand_dim(scal2, "alphaReg", "8", "ii")


def test_lift_alloc_forwarding():
    @proc
    def scal(n: size, alpha: R, x: [R][n]):
        for i in seq(0, n):
            x[i] = alpha * x[i]

    stmt = scal.find("x[_] = _")
    scal1 = divide_loop(scal, "for i in _:_", 8, ("io", "ii"), tail="cut")
    scal2 = bind_expr(scal1, [stmt.rhs().lhs()], "alphaReg")
    scal3 = expand_dim(scal2, "alphaReg", "8", "ii")
    scal4 = lift_alloc(scal3, "alphaReg")

    scal1.forward(stmt)
    scal2.forward(stmt)
    scal3.forward(stmt)
    scal4.forward(stmt)


def test_bind_expr_forwarding(golden):
    @proc
    def scal(n: size, alpha: R, x: [R][n]):
        for i in seq(0, n):
            x[i] = alpha * x[i]

    stmt = scal.find("x[_] = _")
    scal1 = divide_loop(scal, "for i in _:_", 8, ("io", "ii"), tail="cut")
    stmt2 = scal1.find("x[_] = _")
    scal2 = bind_expr(scal1, [stmt.rhs().lhs()], "alphaReg")
    assert str(scal2.forward(stmt)._impl.get_root()) == golden
    assert str(scal2.forward(stmt2)._impl.get_root()) == golden


def test_reorder_loops_forwarding(golden):
    @proc
    def foo():
        for i in seq(0, 4):
            for j in seq(0, 4):
                for k in seq(0, 4):
                    x: i8

    i_loop = foo.find("for i in _:_")
    j_loop = foo.find("for j in _:_")
    k_loop = foo.find("for k in _:_")
    foo = reorder_loops(foo, i_loop)
    foo = reorder_loops(foo, i_loop)
    foo = reorder_loops(foo, j_loop)
    foo = reorder_loops(foo, k_loop)
    assert str(foo) == golden


def test_split_write_forwarding(golden):
    @proc
    def foo(x: i32):
        x += 1 + 2
        x = 3 + 4

    stmt0 = foo.body()[0]
    stmt1 = foo.body()[1]
    body = foo.body()

    foo = split_write(foo, stmt0)
    with pytest.raises(InvalidCursorError, match=""):
        foo.forward(stmt0)
    assert foo.body()[:2] == foo.forward(stmt0.as_block())
    assert foo.body()[2] == foo.forward(stmt1)
    assert foo.body() == foo.forward(body)

    foo = split_write(foo, stmt1)
    with pytest.raises(InvalidCursorError, match=""):
        foo.forward(stmt0.rhs())

    assert foo.body()[:2] == foo.forward(stmt0.as_block())
    assert foo.body()[2:] == foo.forward(stmt1.as_block())
    assert foo.body() == foo.forward(body)


def test_vectorize_forwarding(golden):
    @proc
    def scal(n: size, alpha: R, x: [R][n]):
        for i in seq(0, n):
            x[i] = alpha * x[i]

    stmt = scal.find("x[_] = _")
    scal1 = divide_loop(scal, "for i in _:_", 8, ("io", "ii"), tail="cut")
    scal2 = bind_expr(scal1, [stmt.rhs().lhs()], "alphaReg")
    scal3 = expand_dim(scal2, "alphaReg", "8", "ii")
    scal4 = lift_alloc(scal3, "alphaReg")
    scal5 = fission(scal4, scal4.find("alphaReg[_] = _").after())

    assert str(scal5.forward(stmt)) == golden


def test_unroll_buffer_forwarding(golden):
    @proc
    def foo():
        src: i32[2]
        src[0] = 1.0
        src[1] = 1.0

    assn1 = foo.find("src[_] = _")
    const1 = assn1.rhs()
    assn2 = foo.find("src[_] = _ #1")
    foo1 = unroll_buffer(foo, "src: _", 0)

    tests = [foo1.forward(assn1), foo1.forward(assn2), foo1.forward(const1).parent()]

    assert "\n".join([str(test) for test in tests]) == golden


def test_forwarding_for_procs_with_identical_code():
    @proc
    def foo():
        x: f32[8] @ AVX2
        for i in seq(0, 8):
            x[i] += 1.0

    alloc_cursor = foo.find("x : _")
    foo = set_memory(foo, alloc_cursor, AVX2)
    loop_cursor = foo.find_loop("i")
    foo = expand_dim(foo, alloc_cursor, "1", "0")
    foo.forward(loop_cursor)


def test_delete_pass_forwarding():
    @proc
    def foo(x: R):
        for i in seq(0, 16):
            x = 1.0
            pass
            for j in seq(0, 2):
                pass
                pass
            pass
        x = 0.0

    i_loop = foo.body()[0]
    assign_1 = i_loop.body()[0]
    assign_0 = foo.body()[1]

    foo = delete_pass(foo)
    assert isinstance(foo.forward(i_loop), ForCursor)
    assert isinstance(foo.forward(assign_1), AssignCursor)
    assert isinstance(foo.forward(assign_0), AssignCursor)


def test_extract_subproc_forwarding():
    @proc
    def foo(N: size, M: size, K: size, x: R[N, K + M]):
        assert N >= 8
        x[0, 0] = 0.0
        for i in seq(0, 8):
            x[i, 0] += 2.0

    block = foo.body()
    foo, new = extract_subproc(foo, block, "fooooo")
    block = foo.forward(block)
    assert len(block) == 1
    assert isinstance(block[0], CallCursor)


def test_arg_cursor(golden):
    @proc
    def scal(n: size, alpha: R, x: [R][n, n]):
        for i in seq(0, n):
            x[i, i] = alpha * x[i, i]

    args = scal.args()

    output = ""
    for arg in args:
        output += f"{arg.name()}, {arg.is_tensor()}"
        if arg.is_tensor():
            for dim in arg.shape():
                output += f", {dim._impl._node}"
        output += "\n"

    assert output == golden


def test_argcursor_introspection():
    @proc
    def bar(n: size, x: f32[n], y: [f64][8], result: R):
        pass

    n_arg = bar.args()[0]
    assert isinstance(n_arg, ArgCursor)
    assert n_arg.name() == "n"
    with pytest.raises(AssertionError, match=""):
        mem = n_arg.mem()
    assert n_arg.is_tensor() == False
    with pytest.raises(AssertionError, match=""):
        shape = n_arg.shape()
    assert n_arg.type() == ExoType.Size

    x_arg = bar.args()[1]
    assert isinstance(x_arg, ArgCursor)
    assert x_arg.name() == "x"
    assert x_arg.mem() == DRAM
    assert x_arg.is_tensor() == True
    x_arg_shape = x_arg.shape()
    assert isinstance(x_arg_shape, ExprListCursor)
    assert len(x_arg_shape) == 1
    assert isinstance(x_arg_shape[0], ReadCursor)
    assert x_arg_shape[0].name() == "n"
    assert isinstance(x_arg_shape[0].idx(), ExprListCursor)
    assert len(x_arg_shape[0].idx()) == 0
    assert x_arg.type() == ExoType.F32

    y_arg = bar.args()[2]
    assert isinstance(y_arg, ArgCursor)
    assert y_arg.name() == "y"
    assert y_arg.mem() == DRAM
    assert y_arg.is_tensor() == True
    y_arg_shape = y_arg.shape()
    assert isinstance(y_arg_shape, ExprListCursor)
    assert len(y_arg_shape) == 1
    assert isinstance(y_arg_shape[0], LiteralCursor)
    assert y_arg_shape[0].value() == 8
    assert y_arg.type() == ExoType.F64

    result_arg = bar.args()[3]
    assert isinstance(result_arg, ArgCursor)
    assert result_arg.name() == "result"
    assert result_arg.mem() == DRAM
    assert result_arg.is_tensor() == False
    with pytest.raises(AssertionError, match=""):
        shape = result_arg.shape()
    assert result_arg.type() == ExoType.R


def test_cut_loop_forwarding():
    @proc
    def foo(n: size):
        for i in seq(0, 10):
            pass

    loop_cursor = foo.find_loop("i")
    pass_cursor = loop_cursor.body()[0]
    foo = cut_loop(foo, loop_cursor, 3)
    loop_cursor = foo.forward(loop_cursor)
    second_loop = loop_cursor.next()
    pass_cursor = foo.forward(pass_cursor)

    assert isinstance(loop_cursor, ForCursor)
    assert isinstance(second_loop, ForCursor)
    assert isinstance(pass_cursor, PassCursor)
    assert isinstance(pass_cursor.parent(), ForCursor)
    assert pass_cursor.parent().hi().value() == 3


def test_shift_loop_forwarding():
    @proc
    def foo(x: f32[10]):
        for i in seq(0, 10):
            x[i] = 1.0

    loop_cursor = foo.find_loop("i")
    assign_cursor = loop_cursor.body()[0]
    foo = shift_loop(foo, loop_cursor, 3)
    loop_cursor = foo.forward(loop_cursor)
    assign_cursor = foo.forward(assign_cursor)

    assert isinstance(loop_cursor, ForCursor)
    assert isinstance(assign_cursor, AssignCursor)
    assert isinstance(assign_cursor.parent(), ForCursor)


def test_eliminate_dead_code_forwarding():
    @proc
    def foo():
        x: f32 @ DRAM
        for i in seq(0, 8):
            if i + 3 > -1:
                x = 0.0
                pass
            else:
                x += 1.0
                pass
                pass

    loop_cursor = foo.find_loop("i")
    if_cursor = loop_cursor.body()[0]
    if_true_stmt = if_cursor.body()[0]
    if_false_stmt = if_cursor.orelse()[0]
    foo = eliminate_dead_code(foo, "if _:_ #0")
    loop_cursor = foo.forward(loop_cursor)
    with pytest.raises(InvalidCursorError, match=""):
        if_cursor = foo.forward(if_cursor)
    if_true_stmt = foo.forward(if_true_stmt)
    with pytest.raises(InvalidCursorError, match=""):
        if_false_stmt = foo.forward(if_false_stmt)

    assert isinstance(loop_cursor, ForCursor)
    assert len(loop_cursor.body()) == 2
    assert isinstance(if_true_stmt, AssignCursor)
    assert isinstance(if_true_stmt.parent(), ForCursor)


def test_eliminate_dead_code_forwarding2():
    @proc
    def foo():
        x: f32 @ DRAM
        for i in seq(0, 8):
            if i + 3 < -1:
                x = 0.0
                pass
            else:
                x += 1.0
                pass
                pass

    loop_cursor = foo.find_loop("i")
    if_cursor = loop_cursor.body()[0]
    if_true_stmt = if_cursor.body()[0]
    if_false_stmt = if_cursor.orelse()[0]
    foo = eliminate_dead_code(foo, "if _:_ #0")
    loop_cursor = foo.forward(loop_cursor)
    with pytest.raises(InvalidCursorError, match=""):
        if_cursor = foo.forward(if_cursor)
    with pytest.raises(InvalidCursorError, match=""):
        if_true_stmt = foo.forward(if_true_stmt)
    if_false_stmt = foo.forward(if_false_stmt)

    assert isinstance(loop_cursor, ForCursor)
    assert len(loop_cursor.body()) == 3
    assert isinstance(if_false_stmt, ReduceCursor)
    assert isinstance(if_false_stmt.parent(), ForCursor)


def test_eliminate_dead_code_forwarding3():
    @proc
    def foo():
        x: f32 @ DRAM
        for i in seq(0, 8):
            if i + 3 > -1:
                x = 0.0
                pass

    loop_cursor = foo.find_loop("i")
    if_cursor = loop_cursor.body()[0]
    if_true_stmt = if_cursor.body()[0]
    foo = eliminate_dead_code(foo, "if _:_ #0")
    loop_cursor = foo.forward(loop_cursor)
    with pytest.raises(InvalidCursorError, match=""):
        if_cursor = foo.forward(if_cursor)
    if_true_stmt = foo.forward(if_true_stmt)

    assert isinstance(loop_cursor, ForCursor)
    assert len(loop_cursor.body()) == 2
    assert isinstance(if_true_stmt, AssignCursor)
    assert isinstance(if_true_stmt.parent(), ForCursor)


def test_eliminate_dead_code_forwarding4():
    @proc
    def foo():
        x: f32 @ DRAM
        for i in seq(0, 8):
            if i + 3 < -1:
                x = 0.0
                pass

    loop_cursor = foo.find_loop("i")
    if_cursor = loop_cursor.body()[0]
    if_true_stmt = if_cursor.body()[0]
    foo = eliminate_dead_code(foo, "if _:_ #0")
    loop_cursor = foo.forward(loop_cursor)
    with pytest.raises(InvalidCursorError, match=""):
        if_cursor = foo.forward(if_cursor)
    with pytest.raises(InvalidCursorError, match=""):
        if_true_stmt = foo.forward(if_true_stmt)

    assert isinstance(loop_cursor, ForCursor)
    assert len(loop_cursor.body()) == 1
    assert isinstance(loop_cursor.body()[0], PassCursor)


def test_eliminate_dead_code_forwarding5():
    @proc
    def foo(n: size):
        for i in seq(0, n):
            x: f32

    foo = specialize(foo, foo.find_loop("i"), "0 < n")
    else_loop = foo.find_loop("i #1")
    else_loop_alloc = else_loop.body()[0]
    foo = eliminate_dead_code(foo, else_loop)

    with pytest.raises(InvalidCursorError, match=""):
        foo.forward(else_loop)
    with pytest.raises(InvalidCursorError, match=""):
        foo.forward(else_loop_alloc)


def test_specialize_forwarding():
    @proc
    def foo(n: size, a: f32):
        a = 1.0
        a = 2.0

    body = foo.body()
    foo = specialize(foo, body, ["n > 0"])
    assert foo.forward(body) == foo.body()


def test_match_depth(golden):
    @proc
    def foo(x: i8):
        for i in seq(0, 8):
            if i + 3 < -1:
                x = 0.0
                pass
        for i in seq(0, 2):
            x = 1.0

    c = match_depth(foo.find("x = 1.0"), foo.find_loop("i"))
    assert str(c) == golden


def test_match_depth_fail():
    @proc
    def foo(x: i8):
        for i in seq(0, 8):
            x = 1.0
        for j in seq(0, 2):
            x = 2.0

    @proc
    def bar():
        pass

    with pytest.raises(
        CursorNavigationError,
        match="cursor_to_match's parent is not an ancestor of cursor",
    ):
        match_depth(foo.find("x = _ #0"), foo.find("x = _ #1"))

    with pytest.raises(AssertionError, match="cursors originate from different procs"):
        match_depth(foo.find("x = _"), bar.find("pass"))


def test_get_enclosing_loop_by_name(golden):
    @proc
    def foo(x: i8):
        for i in seq(0, 5):
            for j in seq(0, 5):
                if i == 0:
                    x = 1.0

    c1 = get_enclosing_loop(foo, foo.find("x = _"))
    c2 = get_enclosing_loop_by_name(foo, foo.find("x = _"), "i")

    assert "\n\n".join([str(c) for c in [c1, c2]]) == golden


def test_get_enclosing_loop_by_name_fail():
    @proc
    def foo(x: i8):
        for i in seq(0, 8):
            x = 1.0
        for j in seq(0, 2):
            x = 2.0
        x = 3.0

    with pytest.raises(CursorNavigationError, match="no enclosing loop found"):
        get_enclosing_loop_by_name(foo, foo.find("x = _"), foo.find_loop("j"))


def test_is_ancestor_of_and_lca():
    @proc
    def foo():
        for i in seq(0, 10):
            for j in seq(0, 10):
                pass
            x: i8

    i_loop = foo.find_loop("i")
    j_loop = foo.find_loop("j")
    x_alloc = foo.find("x:_")
    pass_stmt = foo.find("pass")

    assert i_loop.is_ancestor_of(i_loop)

    assert i_loop.is_ancestor_of(j_loop)
    assert i_loop.is_ancestor_of(pass_stmt)
    assert i_loop.is_ancestor_of(x_alloc)

    assert not j_loop.is_ancestor_of(i_loop)
    assert j_loop.is_ancestor_of(pass_stmt)
    assert not j_loop.is_ancestor_of(x_alloc)

    assert not pass_stmt.is_ancestor_of(i_loop)
    assert not pass_stmt.is_ancestor_of(j_loop)

    assert not x_alloc.is_ancestor_of(i_loop)
    assert not x_alloc.is_ancestor_of(j_loop)

    assert get_lca(foo, x_alloc, pass_stmt) == i_loop
    assert get_lca(foo, pass_stmt, x_alloc) == i_loop
    assert get_lca(foo, i_loop, x_alloc) == i_loop
    assert get_lca(foo, x_alloc, i_loop) == i_loop


def test_cursor_find_loop():
    @proc
    def foo(n: size, x: i8[n]):
        for i in seq(0, n):
            pass
        if n > 1:
            for i in seq(0, n):
                x[i] = 0.0

    i_loop2 = foo.find("for i in _:_ #1")
    if_stmt = foo.find("if _: _ ")
    i_loop_alternative = if_stmt.find("for i in _: _")
    assert i_loop2 == i_loop_alternative



from __future__ import annotations

import numpy as np

from exo import proc


# Test 1 is Full 1D convolution
def gen_conv1d():
    @proc
    def conv1d(n: size, m: size, r: size, x: R[n], w: R[m], res: R[r]):
        for i in seq(0, r):
            res[i] = 0.0
        for i in seq(0, r):
            for j in seq(0, n):
                if j < i + 1 and j >= i - (m - 1):
                    res[i] += x[j] * w[i - j]

    return conv1d


def test_conv1d():
    conv1d = gen_conv1d()
    n = 5
    m = 3
    r = n + m - 1
    x = np.array([0.2, 0.5, -0.4, 1.0, 0.0])
    w = np.array([0.6, 1.9, -2.2])
    res = np.random.uniform(size=r)
    conv1d.interpret(n=n, m=m, r=r, x=x, w=w, res=res)
    np.testing.assert_almost_equal(res, [0.12, 0.68, 0.27, -1.26, 2.78, -2.2, 0])


# add vector test
def gen_add_vec():
    @proc
    def add_vec(n: size, x: R[n], y: R[n], res: R[n]):
        for i in seq(0, n):
            res[i] = x[i] + y[i]

    return add_vec


def test_add_vec():
    add_vec = gen_add_vec()
    x = np.array([3.0, 6.0, 9.0])
    y = np.array([1.0, 2.0, 3.0])
    res = np.random.uniform(size=3)
    add_vec.interpret(n=3, x=x, y=y, res=res)
    np.testing.assert_almost_equal(res, [4, 8, 12])


# multiply matrix test
#   C = A * B
def gen_gemm():
    @proc
    def gemm(
        n: size,
        m: size,
        p: size,
        C: R[n, m],
        A: R[n, p],
        B: R[p, m],
    ):
        for i in seq(0, n):
            for j in seq(0, m):
                C[i, j] = 0.0
                for k in seq(0, p):
                    C[i, j] += A[i, k] * B[k, j]

    return gemm


def test_gemm():
    A = np.array([[-1.0, 4.0], [-2.0, 5.0], [6.0, -3.0], [7.0, 8.0]])

    B = np.array([[9.0, 0.0, 2.0], [3.0, 1.0, 10.0]])

    C_answer = [
        [3.0, 4.0, 38.0],
        [-3.0, 5.0, 46.0],
        [45.0, -3.0, -18.0],
        [87.0, 8.0, 94.0],
    ]

    gemm = gen_gemm()
    C = np.random.uniform(size=(4, 3))
    gemm.interpret(n=4, m=3, p=2, A=A, B=B, C=C)
    np.testing.assert_almost_equal(C, C_answer)



from __future__ import annotations

import itertools
import os
import platform

import pytest

from exo import proc
from exo.platforms.rvv import *
from exo.stdlib.scheduling import *
from exo.memory import MemGenError

import numpy as np


@pytest.fixture
def test_rvv_vfmul():
    """
    Compute C[i] = A[i] * B[l]
    """

    @proc
    def vfmul(
        M: size,
        C: f32[M] @ DRAM,
        A: f32[M] @ DRAM,
        B: f32[M] @ DRAM,
    ):
        # pragma: no cover
        assert M == 8
        for i in seq(0, 8):
            C[i] += A[i] * B[i]

    def simple_vfmul(p=vfmul):
        p = divide_loop(p, "i", 4, ["io", "ii"], perfect=True)
        t = "C[4 * io + ii]"
        p = stage_mem(p, "C[_] += _", t, "C_reg")
        p = expand_dim(p, "C_reg", 4, "ii")
        p = expand_dim(p, "C_reg", 2, "io")

        p = lift_alloc(p, "C_reg", n_lifts=2)
        p = autofission(p, p.find("C_reg[_] = _").after(), n_lifts=2)
        p = autofission(p, p.find("C[_] = _").before(), n_lifts=2)

        p = replace(p, "for ii in _: _ #0", rvv_vld_4xf32)
        p = replace(p, "for ii in _: _ #1", rvv_vst_4xf32)

        p = set_memory(p, "C_reg", RVV)
        for buf in ["A", "B"]:
            p = bind_expr(p, f"{buf}[_]", f"{buf}_vec")
            p = expand_dim(p, f"{buf}_vec", 4, "ii")
            p = expand_dim(p, f"{buf}_vec", 2, "io")
            p = lift_alloc(p, f"{buf}_vec", n_lifts=2)
            p = autofission(p, p.find(f"{buf}_vec[_] = _").after(), n_lifts=2)
            p = replace(p, "for ii in _: _ #0", rvv_vld_4xf32)
            p = set_memory(p, f"{buf}_vec", RVV)

        p = replace(p, "for ii in _: _ #0", rvv_vfmacc_4xf32_4xf32)
        return p

    simple_rvv_vfmul = simple_vfmul()

    return simplify(simple_rvv_vfmul)


@pytest.mark.isa("rvv")
def test_gen_rvv_vfmul(golden, test_rvv_vfmul):
    assert str(test_rvv_vfmul) == golden



from __future__ import annotations
from dataclasses import dataclass

from exo import *
from exo.syntax import *
from exo.stdlib.scheduling import *
from exo.API_cursors import *

from .inspection import *


# TODO: delete once this feature is supported by the primitives


@dataclass
class fission_cursors:
    scope1: ForCursor | IfCursor
    scope2: ForCursor | IfCursor

    def __iter__(self):
        yield self.scope1
        yield self.scope2


def fission_(proc, gap, rc=False):
    gap = proc.forward(gap)
    stmt = gap.anchor()
    is_after = stmt.after() is gap
    proc = fission(proc, gap)

    scope1 = proc.forward(stmt).parent()
    scope2 = scope1.next()

    if is_after:
        scope1, scope2 = scope2, scope1

    if rc:
        return proc, fission_cursors(scope1, scope2)
    return proc


@dataclass
class divide_loop_cursors:
    outer_loop: ForCursor
    inner_loop: ForCursor
    tail_loop: ForCursor

    def __iter__(self):
        yield self.outer_loop
        yield self.inner_loop
        yield self.tail_loop


def divide_loop_(proc, loop_cursor, div_const, tail="guard", rc=False):
    loop_cursor = proc.forward(loop_cursor)
    loop_iter = loop_cursor.name()
    perfect = tail == "perfect"
    if tail == "perfect":
        tail = "cut"
        perfect = True
    proc = divide_loop(
        proc,
        loop_cursor,
        div_const,
        (loop_iter + "o", loop_iter + "i"),
        tail=tail,
        perfect=perfect,
    )
    if not rc:
        return proc

    outer_loop = proc.forward(loop_cursor)
    inner_loop = outer_loop.body()[0]

    if perfect == True or tail == "guard":
        tail_loop = InvalidCursor()
    else:
        tail_loop = outer_loop.next()

    outer_loop = proc.forward(outer_loop)
    inner_loop = proc.forward(inner_loop)
    if not isinstance(tail_loop, InvalidCursor):
        tail_loop = proc.forward(tail_loop)
    return proc, divide_loop_cursors(outer_loop, inner_loop, tail_loop)


@dataclass
class stage_mem_cursors:
    alloc: AllocCursor
    load_stage: Cursor
    block: BlockCursor
    store_stage: Cursor

    def __iter__(self):
        yield self.alloc
        yield self.load_stage
        yield self.block
        yield self.store_stage


def stage_mem_(proc, block, buff, new_buff_name, accum=False, rc=False):
    if not isinstance(block, BlockCursor):
        block = proc.forward(block)
        block = block.as_block()

    block_first = block[0]
    block_last = block[-1]
    proc = stage_mem(proc, block, buff, new_buff_name, accum)
    block_first = proc.forward(block_first)
    block_last = proc.forward(block_last)
    alloc = block_first.prev().prev()
    load_stage = block_first.prev()
    block = block_first.as_block().expand(0, len(block) - 1)
    store_stage = block_last.next()
    if not rc:
        return proc
    return proc, stage_mem_cursors(alloc, load_stage, block, store_stage)


@dataclass
class cut_loop_cursors:
    loop1: ForCursor
    loop2: ForCursor

    def __iter__(self):
        yield self.loop1
        yield self.loop2


def cut_loop_(proc, loop, expr, rc=False):
    proc = cut_loop(proc, loop, expr)

    if not rc:
        return proc

    loop1 = proc.forward(loop)
    loop2 = loop1.next()
    return proc, cut_loop_cursors(loop1, loop2)


@dataclass
class specialize_cursors:
    if_stmt: Cursor

    def __iter__(self):
        yield self.if_stmt


def specialize_(proc, stmt, cond, rc=False):
    stmt = proc.forward(stmt)
    parent = stmt.parent()
    index = get_index_in_body(proc, stmt)
    proc = specialize(proc, stmt, cond)
    if not rc:
        return proc
    is_else = False
    if (
        isinstance(parent, IfCursor)
        and not isinstance(parent.orelse(), InvalidCursor)
        and index < len(parent.orelse())
        and parent.orelse()[index] == stmt
    ):
        is_else = True
    if not isinstance(parent, InvalidCursor):
        parent = proc.forward(parent)
    else:
        parent = proc

    if_stmt = parent.body()[index] if not is_else else parent.orelse()[index]
    return proc, specialize_cursors(if_stmt)



import ast as pyast
import inspect
import re
import types
from pathlib import Path
from typing import Optional, Union, List

import exo.LoopIR_scheduling as scheduling
from exo.LoopIR_scheduling import SchedulingError

from .API_types import ProcedureBase, ExoType
from . import LoopIR as LoopIR
from .LoopIR_compiler import run_compile, compile_to_strings
from .LoopIR_interpreter import run_interpreter
from .LoopIR_unification import DoReplace, UnificationError
from .configs import Config
from .effectcheck import InferEffects, CheckEffects
from .memory import Memory
from .parse_fragment import parse_fragment
from .pattern_match import match_pattern
from .prelude import *
from .new_eff import Check_Aliasing

# Moved to new file
from .proc_eqv import decl_new_proc, derive_proc, assert_eqv_proc, check_eqv_proc
from .pyparser import get_ast_from_python, Parser, get_src_locals
from .reflection import LoopIR_to_QAST
from .typecheck import TypeChecker

from . import API_cursors as C
from . import internal_cursors as IC

# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Top-level decorator


def proc(f, _instr=None) -> "Procedure":
    if not isinstance(f, types.FunctionType):
        raise TypeError("@proc decorator must be applied to a function")

    body, getsrcinfo = get_ast_from_python(f)
    assert isinstance(body, pyast.FunctionDef)

    parser = Parser(
        body,
        getsrcinfo,
        func_globals=f.__globals__,
        srclocals=get_src_locals(depth=3 if _instr else 2),
        instr=_instr,
        as_func=True,
    )
    return Procedure(parser.result())


def instr(c_instr, c_global=""):
    if not isinstance(c_instr, str):
        raise TypeError("@instr decorator must be @instr(<your instuction>)")

    def inner(f):
        if not isinstance(f, types.FunctionType):
            raise TypeError("@instr decorator must be applied to a function")

        return proc(f, _instr=(c_instr, c_global))

    return inner


def config(_cls=None, *, readwrite=True):
    def parse_config(cls):
        if not inspect.isclass(cls):
            raise TypeError("@config decorator must be applied to a class")

        body, getsrcinfo = get_ast_from_python(cls)
        assert isinstance(body, pyast.ClassDef)

        parser = Parser(
            body,
            getsrcinfo,
            func_globals={},
            srclocals=get_src_locals(depth=2),
            as_config=True,
        )
        return Config(*parser.result(), not readwrite)

    if _cls is None:
        return parse_config
    else:
        return parse_config(_cls)


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
#   iPython Display Object


class MarkDownBlob:
    def __init__(self, mkdwn_str):
        self.mstr = mkdwn_str

    def _repr_markdown_(self):
        return self.mstr


class FindBefore(LoopIR.LoopIR_Do):
    def __init__(self, proc, stmt):
        self.stmt = stmt
        self.result = None
        super().__init__(proc)

    def result(self):
        return self.result

    def do_stmts(self, stmts):
        prev = None
        for s in stmts:
            if s == self.stmt:
                self.result = prev
                return
            else:
                self.do_s(s)
                prev = s


class FindDup(LoopIR.LoopIR_Do):
    def __init__(self, proc):
        self.result = False
        self.env = []
        super().__init__(proc)

    def result(self):
        return self.result

    def do_s(self, s):
        for e in self.env:
            if s is e:
                self.result = True
                print(s)
        self.env.append(s)

        super().do_s(s)


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
#   Procedure Objects


def compile_procs(proc_list, basedir: Path, c_file: str, h_file: str):
    c_data, h_data = compile_procs_to_strings(proc_list, h_file)
    (basedir / c_file).write_text(c_data)
    (basedir / h_file).write_text(h_data)


def compile_procs_to_strings(proc_list, h_file_name: str):
    assert isinstance(proc_list, list)
    assert all(isinstance(p, Procedure) for p in proc_list)
    return run_compile([p._loopir_proc for p in proc_list], h_file_name)


class Procedure(ProcedureBase):
    def __init__(
        self,
        proc,
        _provenance_eq_Procedure: "Procedure" = None,
        _forward=None,
        _mod_config=None,
    ):
        super().__init__()

        _mod_config = _mod_config or frozenset()

        if isinstance(proc, LoopIR.UAST.proc):
            proc = TypeChecker(proc).get_loopir()
            proc = InferEffects(proc).result()
            CheckEffects(proc)
            Check_Aliasing(proc)

        assert isinstance(proc, LoopIR.LoopIR.proc)

        # add this procedure into the equivalence tracking mechanism
        if _provenance_eq_Procedure:
            derive_proc(
                _provenance_eq_Procedure._loopir_proc, proc, frozenset(_mod_config)
            )
        else:
            decl_new_proc(proc)

        if _forward is None:

            def _forward(_):
                raise NotImplementedError(
                    "This forwarding function has not been implemented"
                )

        self._loopir_proc = proc
        self._provenance_eq_Procedure = _provenance_eq_Procedure
        self._forward = _forward

    def forward(self, cur: C.Cursor):
        p = self
        fwds = []
        while p is not None and p is not cur.proc():
            fwds.append(p._forward)
            p = p._provenance_eq_Procedure

        ir = cur._impl
        for fn in reversed(fwds):
            ir = fn(ir)

        return C.lift_cursor(ir, self)

    def __str__(self):
        return str(self._loopir_proc)

    def __eq__(self, other):
        if not isinstance(other, Procedure):
            return False
        return self._loopir_proc == other._loopir_proc

    def _repr_markdown_(self):
        return "```python\n" + self.__str__() + "\n```"

    def INTERNAL_proc(self):
        return self._loopir_proc

    # -------------------------------- #
    #     introspection operations
    # -------------------------------- #

    def check_effects(self):
        self._loopir_proc = InferEffects(self._loopir_proc).result()
        CheckEffects(self._loopir_proc)
        return self

    def name(self):
        return self._loopir_proc.name

    def show_effects(self):
        self._loopir_proc = InferEffects(self._loopir_proc).result()
        return str(self._loopir_proc.eff)

    def show_effect(self, stmt_pattern):
        self._loopir_proc = InferEffects(self._loopir_proc).result()
        if match := match_pattern(
            self._root(), stmt_pattern, call_depth=1, default_match_no=0
        ):
            assert len(match[0]) == 1, "Must match single statements"
            return str(match[0][0]._node.eff)
        raise SchedulingError("failed to find statement", pattern=stmt_pattern)

    def is_instr(self):
        return self._loopir_proc.instr is not None

    def get_instr(self):
        return self._loopir_proc.instr.c_instr

    def args(self):
        if args := self._root()._child_block("args"):
            return C.lift_cursor(args, self)
        return []

    def body(self):
        """
        Return a BlockCursor selecting the entire body of the Procedure
        """
        block = self._root()._child_block("body")
        return C.lift_cursor(block, self)

    def find(self, pattern, many=False):
        """
        Find the most specific possible cursor for the given pattern.
        For example, a pattern matching a single assignment statement
        will return an AssignCursor, not a StmtCursor or BlockCursor.

        If the optional parameter `many` is set to True, then return a list,
        potentially containing more than one Cursor.

        In any event, if no matches are found, a SchedulingError is raised
        """
        return C.find(self._root(), self, pattern, many)

    def find_loop(self, pattern, many=False):
        """
        This is the same as proc.find(...), except if the supplied pattern
        is of the form 'name' or 'name #n', then it will be auto-expanded
        to 'for name in _:_' or 'for name in _:_ #n'
        """
        if not isinstance(pattern, str):
            raise TypeError("expected a pattern string")

        _name_count_re = r"^([a-zA-Z_]\w*)\s*(\#\s*[0-9]+)?$"
        results = re.search(_name_count_re, pattern)
        if results:
            name, count = results[1], (results[2] if results[2] else "")
            pattern = f"for {name} in _: _{count}"

        return self.find(pattern, many)

    def find_alloc_or_arg(self, pattern):
        _name_count_re = r"^([a-zA-Z_]\w*)\s*(\#\s*[0-9]+)?$"
        results = re.search(_name_count_re, pattern)
        if results:
            name, count = results[1], (results[2] if results[2] else "")
            for arg in self.args():
                if arg._impl._node.name.name() == name:
                    return arg

            pattern = f"{name}: _{count}"

        return self.find(pattern)

    def find_all(self, pattern):
        return self.find(pattern, many=True)

    def get_ast(self, pattern=None):
        if pattern is None:
            return LoopIR_to_QAST(self._loopir_proc).result()

        # do pattern matching
        match = match_pattern(self._root(), pattern, call_depth=1)

        # convert matched sub-trees to QAST
        assert isinstance(match, list)
        if not match:
            return None

        return [
            LoopIR_to_QAST(node._node).result() for block in match for node in block
        ]

    # ---------------------------------------------- #
    #     execution / interpretation operations
    # ---------------------------------------------- #

    def show_c_code(self):
        return MarkDownBlob("```c\n" + self.c_code_str() + "\n```")

    def c_code_str(self):
        decls, defns = compile_to_strings("c_code_str", [self._loopir_proc])
        return decls + "\n" + defns

    def compile_c(self, directory: Path, filename: str):
        compile_procs([self], directory, f"{filename}.c", f"{filename}.h")

    def interpret(self, **kwargs):
        run_interpreter(self._loopir_proc, kwargs)

    # ------------------------------- #
    #     scheduling operations
    # ------------------------------- #

    def has_dup(self):
        """
        Internal check to see if there are any reference diamonds in the AST
        """
        return FindDup(self._loopir_proc).result

    def unsafe_assert_eq(self, other_proc):
        if not isinstance(other_proc, Procedure):
            raise TypeError("expected a procedure as argument")
        assert_eqv_proc(self._loopir_proc, other_proc._loopir_proc)
        return self

    def partial_eval(self, *args, **kwargs):
        if kwargs and args:
            raise ValueError("Must provide EITHER ordered OR named arguments")
        if not kwargs and not args:
            # Nothing to do if empty partial eval
            return self

        p = self._loopir_proc

        if args:
            if len(args) > len(p.args):
                raise TypeError(
                    f"expected no more than {len(p.args)} "
                    f"arguments, but got {len(args)}"
                )
            kwargs = {arg.name: val for arg, val in zip(p.args, args)}
        else:
            # Get the symbols corresponding to the names
            params_map = {sym.name.name(): sym.name for sym in p.args}
            kwargs = {params_map[k]: v for k, v in kwargs.items()}

        p = scheduling.DoPartialEval(kwargs).apply_proc(p)
        return Procedure(p)  # No provenance because signature changed

    def transpose(self, arg_cursor):
        if not (isinstance(arg_cursor, C.ArgCursor) and len(arg_cursor.shape()) == 2):
            raise TypeError("expected a 2D argument cursor")

        ir, _ = scheduling.DoRearrangeDim(arg_cursor._impl, [1, 0])
        return Procedure(ir)  # No provenance because signature changed

    def add_assertion(self, assertion, configs=None):
        if not isinstance(assertion, str):
            raise TypeError("assertion must be an Exo string")

        configs = configs or []

        p = self._loopir_proc
        assertion = parse_fragment(p, assertion, p.body[0], configs=configs)
        p = LoopIR.LoopIR.proc(
            p.name, p.args, p.preds + [assertion], p.body, p.instr, p.eff, p.srcinfo
        )
        return Procedure(p, _provenance_eq_Procedure=None)

    def is_eq(self, proc: "Procedure"):
        eqv_set = check_eqv_proc(self._loopir_proc, proc._loopir_proc)
        return eqv_set == frozenset()

    def _root(self):
        return IC.Cursor.create(self._loopir_proc)



from .API import (
    Procedure,
    compile_procs,
    compile_procs_to_strings,
    proc,
    instr,
    config,
    ExoType,
)
from .LoopIR_scheduling import SchedulingError
from .parse_fragment import ParseFragmentError
from .configs import Config
from .memory import Memory, DRAM
from . import query_asts as QAST

from . import stdlib

__version__ = "0.2.0"

__all__ = [
    "Procedure",
    "compile_procs",
    "compile_procs_to_strings",
    "proc",
    "instr",
    "config",
    "Config",
    "Memory",
    "DRAM",
    "QAST",
    "SchedulingError",
    "ParseFragmentError",
    #
    "stdlib",
    "ExoType",
]



from __future__ import annotations

from exo import Memory, DRAM, instr


def _is_const_size(sz, c):
    return sz.isdecimal() and int(sz) == c


def _is_some_const_size(sz):
    return sz.isdecimal() and int(sz) > 0


# --------------------------------------------------------------------------- #
#   Neon registers
# --------------------------------------------------------------------------- #


class RVV(Memory):
    @classmethod
    def global_(cls):
        return "#include <riscv_vector.h>"

    @classmethod
    def can_read(cls):
        return False

    @classmethod
    def alloc(cls, new_name, prim_type, shape, srcinfo):
        if not shape:
            raise MemGenError(f"{srcinfo}: RVV vectors are not scalar values")

        vec_types = {
            "float": (4, "vfloat32m1_t")
        }  # , "double": (2, "float64x2_t"), "_Float16" : (8, "float16x8_t")}

        if not prim_type in vec_types.keys():
            raise MemGenError(f"{srcinfo}: RVV vectors must be f32 (for now)")

        reg_width, C_reg_type_name = vec_types[prim_type]

        if not _is_const_size(shape[-1], reg_width):

            # This will help with dynamic lengths (I hope)
            if int(shape[-1]) > reg_width:
                raise MemGenError(
                    f"{srcinfo}: RVV vectors of type {prim_type} must be {reg_width}-wide, got {shape}"
                )
        shape = shape[:-1]
        if shape:
            if not all(_is_some_const_size(s) for s in shape):
                raise MemGenError(
                    f"{srcinfo}: Cannot allocate variable numbers of RVV vectors"
                )
            result = f'{C_reg_type_name} {new_name}[{"][".join(map(str, shape))}];'
        else:
            result = f"{C_reg_type_name} {new_name};"

        return result

    @classmethod
    def free(cls, new_name, prim_type, shape, srcinfo):
        return ""

    @classmethod
    def window(cls, basetyp, baseptr, indices, strides, srcinfo):
        assert strides[-1] == "1"
        idxs = indices[:-1] or ""
        if idxs:
            idxs = "[" + "][".join(idxs) + "]"
        return f"{baseptr}{idxs}"


# --------------------------------------------------------------------------- #
#   f32 RVV intrinsics
# --------------------------------------------------------------------------- #

#
# Load, Store, Broadcast, FMAdd, Mul, Add?
#
# float32


@instr("{dst_data} = __riscv_vle32_v_f32m1(&{src_data},{vl});")
def rvv_vld_4xf32(dst: [f32][4] @ RVV, src: [f32][4] @ DRAM, vl: size):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1
    assert vl >= 0
    assert vl <= 4

    for i in seq(0, vl):
        dst[i] = src[i]


@instr("__riscv_vse32_v_f32m1(&{dst_data}, {src_data},{vl});")
def rvv_vst_4xf32(dst: [f32][4] @ DRAM, src: [f32][4] @ RVV, vl: size):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1
    assert vl >= 0
    assert vl <= 4

    for i in seq(0, vl):
        dst[i] = src[i]


@instr("{dst_data} = __riscv_vfmv_v_f_f32m1({src_data},{vl});")
def rvv_broadcast_4xf32(dst: [f32][4] @ RVV, src: [f32][1] @ DRAM, vl: size):
    assert stride(dst, 0) == 1
    assert vl >= 0
    assert vl <= 4

    for i in seq(0, vl):
        dst[i] = src[0]


@instr("{dst_data} = __riscv_vfmv_v_f_f32m1({src_data},{vl});")
def rvv_broadcast_4xf32_scalar(dst: [f32][4] @ RVV, src: f32 @ DRAM, vl: size):
    assert stride(dst, 0) == 1
    assert vl >= 0
    assert vl <= 4

    for i in seq(0, vl):
        dst[i] = src


@instr("{dst_data} = __riscv_vfmv_v_f_f32m1(0.0f,{vl});")
def rvv_broadcast_4xf32_0(dst: [f32][4] @ RVV, vl: size):
    assert stride(dst, 0) == 1
    assert vl >= 0
    assert vl <= 4

    for i in seq(0, vl):
        dst[i] = 0.0


@instr("{dst_data} = __riscv_vfmacc_vv_f32m1({dst_data}, {lhs_data}, {rhs_data},{vl});")
def rvv_vfmacc_4xf32_4xf32(
    dst: [f32][4] @ RVV, lhs: [f32][4] @ RVV, rhs: [f32][4] @ RVV, vl: size
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1
    assert vl >= 0
    assert vl <= 4

    for i in seq(0, vl):
        dst[i] += lhs[i] * rhs[i]


@instr("{dst_data} = __riscv_vfmacc_vf_f32m1{dst_data}, {rhs_data}, {lhs_data},{vl});")
def rvv_vfmacc_4xf32_1xf32(
    dst: [f32][4] @ RVV, lhs: [f32][4] @ RVV, rhs: [f32][1] @ DRAM, vl: size
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1
    assert vl >= 0
    assert vl <= 4

    for i in seq(0, vl):
        dst[i] += lhs[i] * rhs[0]


@instr("{dst_data} = __riscv_vfmacc_vf_f32m1{dst_data}, {lhs_data}, {rhs_data},{vl});")
def rvv_vfmacc_1xf32_4xf32(
    dst: [f32][4] @ RVV, lhs: [f32][1] @ DRAM, rhs: [f32][4] @ RVV, vl: size
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1
    assert vl >= 0
    assert vl <= 4

    for i in seq(0, vl):
        dst[i] += lhs[0] * rhs[i]
